==> libgii_0.8.5.orig.tar.gz: libgii-0.8.5.orig/gg/cpuid.c <==
/* $Id: cpuid.c,v 1.7.2.1 2003/04/14 20:54:22 cegger Exp $
***************************************************************************

   LibGG - General runtime CPU-type detection code.

   Copyright (C) 2002  Brian S. Julin   [bri@tull.umassp.edu]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#include <ggi/gg.h>
#include "config.h"

/* State of CPU detection/selection is kept here. */
static int cpuid_tested = 0;
enum gg_swartype swars_detected = 0;
enum gg_swartype swars_enabled = GG_SWAR_ALL;

#ifdef  COMPILING_FOR_SPARC64
#define HAVE_THIS_ARCH
#define HAVE_SWARTYPE_IMPLEMENTATION

/* If I read things right, all sparc64 have VIS. */
enum gg_swartype ggGetSwarType(void) {
	cpuid_tested = 1;
	return (swars_enabled & (GG_SWAR_VIS | GG_SWAR_64BITC | GG_SWAR_NONE));
}

#endif

#ifdef COMPILING_FOR_ALPHA
#define HAVE_THIS_ARCH

#ifndef CC_CAN_AMASK
#warning Compiler must be able to test CPU capabilities with AMASK for SWAR.
#else

#define HAVE_SWARTYPE_IMPLEMENTATION

enum gg_swartype ggGetSwarType(void) {
	if (!cpuid_tested) {
		unsigned long rvb, rc;

		cpuid_tested = 1;
		rvb = 0x0000000000000100; /* Bit 8 = MVI (a.k.a. MAX) */
		__asm__ __volatile__("amask %1, %0" : "=r" (rc) : "rI" (rvb));
		/* will clear bits representing features that ARE present. */
		if (!rc) swars_detected |= GG_SWAR_MVI;
	}
	return (swars_enabled & 
		(swars_detected | GG_SWAR_64BITC | GG_SWAR_NONE));
}

#endif
#endif


#ifdef	COMPILING_FOR_X86
#define HAVE_THIS_ARCH

#ifndef CC_CAN_CPUID
#warning You must implement CPU detection for this compiler or OS to use SWAR
#else
#define HAVE_SWARTYPE_IMPLEMENTATION

/*********************************** x86 **********************************
   The following section of code was taken from, and may be updated from
   the cpuid system utility program.  It has been stripped of stdio
   functions, human readable strings, and some functions not directly relevant 
   to determining the type of SWAR (SIMD) extensions available.  Some of the
   structure has been left in, although unused, in case it is desirable
   in the future to differentiate older CPUs somehow.

   The original copyright of the cpuid program reads as follows:

   * Intel and AMD x86 CPUID display program v 3.3 (1 Jan 2002)
   * Copyright 2002 Phil Karn, KA9Q
   * Updated 24 Apr 2001 to latest Intel CPUID spec
   * Updated 22 Dec 2001 to decode Intel flag 28, hyper threading
   * Updated 1 Jan 2002 to cover AMD Duron, Athlon
   * May be used under the terms of the GNU Public License (GPL)

   * Reference documents:
   * ftp://download.intel.com/design/pro/applnots/24161809.pdf  (AP-485)
   * http://developer.intel.com/design/Pentium4/manuals/24547103.pdf
   * http://developer.intel.com/design/pentiumiii/applnots/24512501.pdf (AP-909)
   * http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/20734.pdf
   * 
   */

void dointel(int),doamd(int),docyrix(int);

/* Below I do not know why I cannot add edx and ecx into the clobber list,
 * instead of making them fake inputs.  Kinda weird.  Also note normally
 * you should NOT have two asm statements in a row like this because GCC
 * may reorder them.  In this case, it reordering doesn't matter.
 */
#define cpuid(in,a,b,c,d) do { \
__asm__ __volatile__ ("push %%ebx\t\ncpuid\t\npop %%ebx" :                    \
                      "=a" (a), "=c" (c), "=d" (d) :                          \
                      "a" (in), "d" (0), "c" (0): "cc");                      \
__asm__ __volatile__ ("push %%ebx\t\ncpuid\t\npush %%ebx\t\n"                 \
		      "pop %%ecx\t\npop %%ebx" :                              \
                      "=a" (a), "=c" (b) : "a" (in) :                         \
                      "edx", "cc");                                           \
} while(0);

static void x86cpuid(void)
{
	unsigned int i;
	unsigned long li,maxi,maxei,ebx,ecx,edx,unused;

	/* Insert code here to test if CPUID instruction is available */

	/* Dump all the CPUID results in raw hex */
	cpuid(0,maxi,unused,unused,unused);
	maxi &= 0xffff; /* The high-order word is non-zero on some Cyrix CPUs */
	for(i=0;i<=maxi;i++){
		unsigned long eax,ebx,ecx,edx;
		cpuid(i,eax,ebx,ecx,edx);
	}	/* for */

	cpuid (0x80000000,maxei,unused,unused,unused);
	for (li=0x80000000;li<=maxei;li++){
		unsigned long eax,ebx,ecx,edx;
		cpuid(li,eax,ebx,ecx,edx);
	}	/* for */

	/* Vendor ID and max CPUID level supported */
	cpuid(0,unused,ebx,ecx,edx);

	switch(ebx){
	case 0x756e6547: /* Intel */
		dointel(maxi);
		break;
	case 0x68747541: /* AMD */
		doamd(maxi);
		break;
	case 0x69727943: /* Cyrix */
		docyrix(maxi);
		break;
	default:
		break;
	}	/* switch */

	return;
}	/* x86cpuid */


/* Intel-specific information */
void dointel(int maxi)
{
	if(maxi >= 1){
		/* Family/model/type etc */
		int clf,apic_id,feature_flags;
		int extended_model = -1,extended_family = -1;
		unsigned long eax,ebx,edx,unused;
		int stepping,model,family,type,reserved,brand,siblings;

		cpuid(1,eax,ebx,unused,edx);
		stepping = eax & 0xf;
		model = (eax >> 4) & 0xf;
		family = (eax >> 8) & 0xf;
		type = (eax >> 12) & 0x3;
		reserved = eax >> 14;
		clf = (ebx >> 8) & 0xff;
		apic_id = (ebx >> 24) & 0xff;
		siblings = (ebx >> 16) & 0xff;
		feature_flags = edx;

		switch(family) { /* switch kept in case we want it */
		case 3:
			/* i386 */
			break;
		case 4:
			/* i486 */
			break;
		case 5:
			/* Pentium */
			break;
		case 6:
			/* Pentium Pro */
			break;
		case 15:
			/* Pentium 4 */
			break;
		}	/* switch */

		if(family == 15){
			extended_family = (eax >> 20) & 0xff;
		}	/* if */

		switch(family){
		case 3:
			break;
		case 4:
			switch(model){
			case 0:
			case 1:
				/* DX */
				break;
			case 2:
				/* SX */
				break;
			case 3:
				/* 487/DX2 */
				break;
			case 4:
				/* SL */
				break;
			case 5:
				/* SX2 */
				break;
			case 7:
				/* write-back enhanced DX2 */
				break;
			case 8:
				/* DX4 */
				break;
			}	/* switch */
			break;
		case 5:
			switch(model) {
			case 1:
				/* 60/66 */
				break;
			case 2:
				/* 75-200 */
				break;
			case 3:
				/* for 486 system */
				break;
			case 4:
				swars_detected |= GG_SWAR_MMX;
				break;
			}	/* switch */
			break;
		case 6:
			switch(model) {
			case 1:
				/* Pentium Pro */
				break;
			case 3:
				/* Pentium II Model 3 */
				break;
			case 5:
				/* Pentium II Model 5/Xeon/Celeron */
				break;
			case 6:
				/* Celeron */
				break;
			case 7:
				/* Pentium III/Pentium III Xeon
				 * - external L2 cache */
				break;
			case 8:
				/* Pentium III/Pentium III Xeon
				 * - internal L2 cache */
				break;
			}	/* switch */
			break;
		case 15:
			break;
		}	/* switch */

		if (model == 15){
			extended_model = (eax >> 16) & 0xf;
		}	/* if */

		brand = ebx & 0xff;
		cpuid(0x80000000,eax,ebx,unused,edx);
		if(eax & 0x80000000) {
			/* Extended feature/signature bits supported */
			unsigned int maxe = eax;
			if (maxe >= 0x80000004) {
				unsigned int i;

				for (i=0x80000002; i <= 0x80000004;i++) {
					unsigned long eax,ebx,ecx,edx;

					cpuid(i,eax,ebx,ecx,edx);	
				}	/* for */
			}	/* if */
		}	/* if */

		/*     if(clf)  CLFLUSH instruction cache line size: %d\n */
		/*     if(apic_id)Initial APIC ID: %d\n */

#if 0
		if (feature_flags & (1<<28)) {
			Hyper threading siblings, number is in siblings 
		}	/* if */
#endif

		/* Feature flags */
		if (feature_flags & (1<<23)) swars_detected |= GG_SWAR_MMX;

		/* 24  "FXSR   Fast FP/MMX Streaming SIMD
		 * Extensions save/restore" */
		if (feature_flags & (1<<25)) swars_detected |= GG_SWAR_SSE;
		if (feature_flags & (1<<26)) swars_detected |= GG_SWAR_SSE2;

	}	/* if */
}	/* dointel */



/* AMD-specific information */
void doamd(int maxi)
{
	unsigned long maxei,unused;
	int family = 0;

	/* Do standard stuff */
	if (maxi >= 1){
		unsigned long eax,ebx,edx,unused;
		int stepping,model,reserved;

		cpuid(1,eax,ebx,unused,edx);
		stepping = eax & 0xf;
		model = (eax >> 4) & 0xf;
		family = (eax >> 8) & 0xf;
		reserved = eax >> 12;

		switch(family){
		case 4:
			/* 486 model (number in variable model) */
			break;
		case 5:
			switch (model) {
			case 0:
			case 1:
			case 2:
			case 3:
			case 6:
			case 7:
				/* K6 Model (number in model) */
				break;
			case 8:
				/* K6-2 Model 8 */
				break;
			case 9:
				/* K6-III Model 9 */
				break;
			default:
				/* K5/K6 model (number in model) */
				break;
			}	/* switch */
			break;
		case 6:
			switch (model) {
			case 1:
			case 2:
			case 4:
				/* Athlon model (number in model) */
				break;
			case 3:
				/* Duron model 3 */
				break;
			case 6:
				/* Athlon MP/Mobile Athlon model 6 */
				break;
			case 7:
				/* Mobile Duron Model 7 */
				break;
			default:
				/* Duron/Athlon model (number in model) */
				break;
			}	/* switch */

			break;
		}	/* switch */

		{
		int i;

		for (i=0;i<32;i++) {
			if(family == 5 && model == 0) {
#if 0
			if (i == 9) Global Paging Extensions\n 
			else if(i == 13) 13 - reserved\n
#endif
			} else {
				if(edx & (1<<i)) {
					switch (i) {
					case 22: 
						/* Extra MMX instructions
						 * will have flag elsewhere
						 * later.
						 */
						break;
					case 23:
						swars_detected |= GG_SWAR_MMX;
						break;
					case 24:
						/* "FXSAVE/FXRSTOR" */
						break;
					case 30:
						swars_detected |= GG_SWAR_3DNOW;
						swars_detected |= GG_SWAR_MMX;
						break;
					case 31:
						swars_detected |= GG_SWAR_ADV3DNOW;
						swars_detected |= GG_SWAR_3DNOW;
						swars_detected |= GG_SWAR_MMX;
						break;
					}	/* switch */
				}	/* if */
			}	/* if */
		}	/* for */
	}	/* if */
	}

	/* Check for presence of extended info */
	cpuid(0x80000000,maxei,unused,unused,unused);
	if(maxei == 0)
		return;

	if (maxei >= 0x80000001) {
		unsigned long eax,ebx,ecx,edx;
		int stepping,model,generation,reserved;
		int i;

		cpuid(0x80000001,eax,ebx,ecx,edx);
		stepping = eax & 0xf;
		model = (eax >> 4) & 0xf;
		generation = (eax >> 8) & 0xf;
		reserved = eax >> 12;

		for (i=0;i<32;i++){
			if (family == 5 && model == 0 && i == 9) {
				/* Global Paging Extensions\n */
			} else {
				if (edx & (1<<i)) {
					switch (i) {
					case 22: 
						/* Extra MMX instructions
						 * will have flag elsewhere
						 * later. */
						break;
					case 23:
						swars_detected |= GG_SWAR_MMX;
						break;
					case 24:
						/* "FXSAVE/FXRSTOR" */
						break;
					case 30:
						swars_detected |= GG_SWAR_3DNOW;
						swars_detected |= GG_SWAR_MMX;
						break;
					case 31:
						swars_detected |= GG_SWAR_ADV3DNOW;
						swars_detected |= GG_SWAR_3DNOW;
						swars_detected |= GG_SWAR_MMX;
						break;
					}	/* switch */
				}	/* if */
			}	/* if */
		}	/* for */
	}	/* if */
}	/* doamd */



/* Cyrix-specific information */
void docyrix(int maxi)
{
	unsigned long maxei,unused;
	unsigned int i;

	cpuid(0x80000000,maxei,unused,unused,unused);

	/* Dump extended info, if any, in raw hex */
	for (i = 0x80000000; i <= maxei; i++) {
		unsigned long eax,ebx,ecx,edx;

		cpuid(i,eax,ebx,ecx,edx);
	}	/* for */

	/* Do standard stuff */
	if (maxi >= 1) {
		unsigned long eax,unused,edx;
		int stepping,model,family,reserved;

		cpuid(1,eax,unused,unused,edx);
		stepping = eax & 0xf;
		model = (eax >> 4) & 0xf;
		family = (eax >> 8) & 0xf;
		reserved = eax >> 12;

		switch(family){
		case 4:
			switch (model) {
			case 4:
				/* MediaGX */
				break;
			}	/* switch */
			break;
		case 5:
			switch (model) {
			case 2:
				/* 6x86 */
				break;
			case 4:
				/* BXm */
				break;
			}	/* switch */
			break;
		case 6:
			switch (model) {
			case 0:
				/* 6x86/MX */
				break;
			}	/* switch */
			break;
		}	/* switch */

		if (family == 5 && model == 0) {
			int i;

			for(i=0;i<32;i++) {
				if (edx & (1<<i)) {
					switch (i) {
					case 8:
						/* "COMPXCHG8B Instruction" */
						break;
					case 15:
						/* "CMOV  Conditional
						 * Move Instruction" */
						break;
					case 23:
						swars_detected |= GG_SWAR_MMX;
						break;
					}	/* switch */
				}	/* if */
			}	/* for */
		} else {
			int i;

			for (i=0;i<32;i++) {
				if (edx & (1<<i)) {
					switch (i) {
					case 8:
						/* "COMPXCHG8B Instruction" */
						break;
					case 15:
						/* "CMOV  Conditional
						 * Move Instruction" */
						break;
					case 23:
						swars_detected |= GG_SWAR_MMX;
						break;
					}	/* switch */
				}	/* if */
			}	/* for */
		}	/* if */
	}	/* if */

	/* Check for presence of extended info */
	if(maxei < 0x80000000)
		return;

	if(maxei >= 0x80000001){
		unsigned long eax,ebx,ecx,edx;
		int stepping,model,family,reserved,i;

		cpuid(0x80000001,eax,ebx,ecx,edx);
		stepping = eax & 0xf;
		model = (eax >> 4) & 0xf;
		family = (eax >> 8) & 0xf;
		reserved = eax >> 12;

		switch(family){
		case 4:
			/* MediaGX */
			break;
		case 5:
			/* 6x86/GXm */
			break;
		case 6:
			/* 6x86/MX */
			break;
		}	/* switch */

	 	for (i=0;i<32;i++) {
			if (edx & (1<<i)) {
				switch (i) {
				case 8:
					/* "COMPXCHG8B Instruction" */
					break;
				case 15:
					/* "CMOV  Conditional
					 * Move Instruction" */
					break;
				case 16:
					/* "FPU CMOV" */
					break;
				case 23:
					swars_detected |= GG_SWAR_MMX;
					break;
				case 24:
					swars_detected |= GG_SWAR_MMXPLUS;
					break;
				case 31:
					swars_detected |= GG_SWAR_3DNOW;
					break;
				}	/* switch */
			}	/* if */
		}	/* for */
	}	/* if */
}	/* docyrix */


enum gg_swartype ggGetSwarType(void)
{
	if (!cpuid_tested) x86cpuid();
	cpuid_tested = 1;
	return (swars_enabled & (swars_detected | GG_SWAR_NONE));
}	/* ggGetSwarType */


#endif

#endif

#ifndef HAVE_THIS_ARCH
#warning You must implement cpuid for this architecture to use SWAR.
#endif

#ifndef HAVE_SWARTYPE_IMPLEMENTATION
enum gg_swartype ggGetSwarType(void)
{ 
	/* By default we assume the C SWARs all are faster if the arch
	 * has the types. This may not be true everywhere, but those archs 
	 * can define their own implementation.
	 */
	return GG_SWAR_NONE | GG_SWAR_32BITC /* This needs ifdeffing too */

# ifdef GG_HAVE_INT64
	| GG_SWAR_64BITC
# endif
	;
};
#endif

==> libgii_0.9.1.orig.tar.gz: libgii-0.9.1/gg/cpuid.c <==
/* $Id$
***************************************************************************

   LibGG - General runtime CPU-type detection code.

   Copyright (C) 2002  Brian S. Julin   [bri@tull.umassp.edu]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#include "cpuid.h"

#ifdef HAVE_NO_CPUID
#warning You must implement cpuid for this architecture to use SWAR.
#endif

#ifdef HAVE_NO_SWARTYPE_IMPLEMENTATION
/* State of CPU detection/selection is kept here. */
static int cpuid_tested = 0;
enum gg_swartype swars_detected = 0;
enum gg_swartype swars_enabled = GG_SWAR_ALL;


enum gg_swartype ggGetSwarType(void)
{
	/* By default we assume the C SWARs all are faster if the arch
	 * has the types. This may not be true everywhere, but those archs 
	 * can define their own implementation.
	 */
	return GG_SWAR_NONE | GG_SWAR_32BITC	/* This needs ifdeffing too */
#if 0				/* Better assume to have 32bit machines by default. */
# ifdef GG_HAVE_INT64
	    | GG_SWAR_64BITC
# endif
#endif
	    ;
}
#endif
